#!/usr/bin/python

from StringIO import StringIO
from parserconst import *

def parse_message(pkt):
    return MessageParser().parse(pkt)

class MessageParser(object):
    def parse(self, pkt):
        try:
            return self._parse(pkt)
        except (ValueError, TypeError):
            raise ValueError, "Invalid message"
    def _read_name(self, f, pkt):
        retval = []
        ll = ord(f.read(1))  # label length
        if ll and ll >> 6 == 0:    # regular name
            while ll:
                lab = f.read(ll)
                if len(lab) != ll:
                    raise ValueError
                retval.append(lab)
                ll = ord(f.read(1))  # label length
        elif ll and ll >> 6 == 3:
            # TODO - test this
            offset = (ll & 0x3f) << 8 | ord(f.read(1))
            ff = StringIO(pkt).seek(offset)
            ll = ord(ff.read(1))  # label length
            while ll:
                lab = ff.read(ll)
                if len(lab) != ll:
                    raise ValueError
                retval.append(lab)
                ll = ord(ff.read(1))  # label length
            del ff, offset
        elif ll == 0:
            pass
        else:
            raise ValueError, "Invalid/Unrecognized query"
        return tuple(retval)
    
    def _read_rr(self, f, pkt):
        NAME = self._read_name(f, pkt)
        TYPE = ord(f.read(1)) << 8 | ord(f.read(1))
        CLASS = ord(f.read(1)) << 8 | ord(f.read(1))
        TTL = ord(f.read(1)) << 24L | ord(f.read(1)) << 16 | \
              ord(f.read(1)) << 8 | ord(f.read(1))
        RDLENGTH = ord(f.read(1)) << 8 | ord(f.read(1))
        RDATA = f.read(RDLENGTH)
        if len(RDATA) != RDLENGTH:
            raise ValueError
        return (NAME, TYPE, CLASS, TTL, RDLENGTH, RDATA)
        
    def _parse(self, pkt):
        f = StringIO(pkt)
        
        # Header section
        hdr = [ord(x) for x in f.read(12)]
        ID = hdr[0] << 8 | hdr[1]
        QR = hdr[2] >> 7
        OPCODE = hdr[2] >> 3 & 0x0f
        AA = hdr[2] >> 2 & 1
        TC = hdr[2] >> 1 & 1
        RD = hdr[2] & 1
        RA = hdr[3] >> 7
        Z = hdr[3] >> 4 & 0xf
        RCODE = hdr[3] & 0xf
        QDCOUNT = hdr[4] << 8 | hdr[5]
        ANCOUNT = hdr[6] << 8 | hdr[7]
        NSCOUNT = hdr[8] << 8 | hdr[9]
        ARCOUNT = hdr[10] << 8 | hdr[11]
        
        HEADER = (ID, QR, OPCODE, AA, TC, RD, RA, Z, RCODE,
                  QDCOUNT, ANCOUNT, NSCOUNT, ARCOUNT)
        
        # Note that we don't need sanity checking on the ??COUNT values,
        # because the StringIO will simply return "" when we reach the end of
        # the string, causing a ValueError exception to be raised when it's
        # passed to the ord() function.  In special cases, we raise ValueError
        # ourselves.
        # FIXME - make the above make sense
        
        # Question section
        QUESTION = []
        for i in range(QDCOUNT):
            QNAME = self._read_name(f, pkt)
            QTYPE = ord(f.read(1)) << 8 | ord(f.read(1))
            QCLASS = ord(f.read(1)) << 8 | ord(f.read(1))
            QUESTION.append((QNAME, QTYPE, QCLASS))
        QUESTION = tuple(QUESTION) 

        # Answer section
        ANSWER = []
        for i in range(ANCOUNT):
            ANSWER.append(self._read_rr(f, pkt))
        ANSWER = tuple(ANSWER) 
        
        # Authority section
        AUTHORITY = []
        for i in range(NSCOUNT):
            AUTHORITY.append(self._read_rr(f, pkt))
        AUTHORITY = tuple(AUTHORITY)
        
        # Additional section
        ADDITIONAL = []
        for i in range(ARCOUNT):
            ADDITIONAL.append(self._read_rr(f, pkt))
        ADDITIONAL = tuple(ADDITIONAL)

        return (HEADER, QUESTION, ANSWER, AUTHORITY, ADDITIONAL)


def pretty_format(msg):
    rv = []
    rv.append("HEADER.ID      = %d\n" % msg[HEADER][ID])
    rv.append("HEADER.QR      = %d\n" % msg[HEADER][QR])
    rv.append("HEADER.OPCODE  = %d\n" % msg[HEADER][OPCODE])
    rv.append("HEADER.AA      = %d\n" % msg[HEADER][AA])
    rv.append("HEADER.TC      = %d\n" % msg[HEADER][TC])
    rv.append("HEADER.RD      = %d\n" % msg[HEADER][RD])
    rv.append("HEADER.RA      = %d\n" % msg[HEADER][RA])
    rv.append("HEADER.Z       = %d\n" % msg[HEADER][Z])
    rv.append("HEADER.RCODE   = %d\n" % msg[HEADER][RCODE])
    rv.append("HEADER.QDCOUNT = %d\n" % msg[HEADER][QDCOUNT])
    rv.append("HEADER.ANCOUNT = %d\n" % msg[HEADER][ANCOUNT])
    rv.append("HEADER.NSCOUNT = %d\n" % msg[HEADER][NSCOUNT])
    rv.append("HEADER.ARCOUNT = %d\n" % msg[HEADER][ARCOUNT])
    for i, x in enumerate(msg[QUESTION]):
        rv.append("QUESTION[%d].QNAME  = %s\n" % (i, `x[QNAME]`))
        rv.append("QUESTION[%d].QTYPE  = %d\n" % (i, x[QTYPE]))
        rv.append("QUESTION[%d].QCLASS = %d\n" % (i, x[QCLASS]))
    for i, x in enumerate(msg[ANSWER]):
        rv.append("ANSWER[%d].NAME     = %s\n" % (i, `x[NAME]`))
        rv.append("ANSWER[%d].TYPE     = %d\n" % (i, x[TYPE]))
        rv.append("ANSWER[%d].CLASS    = %d\n" % (i, x[CLASS]))
        rv.append("ANSWER[%d].TTL      = %d\n" % (i, x[TTL]))
        rv.append("ANSWER[%d].RDLENGTH = %d\n" % (i, x[RDLENGTH]))
        rv.append("ANSWER[%d].RDATA    = %s\n" % (i, `x[RDATA]`))
    for i, x in enumerate(msg[AUTHORITY]):
        rv.append("AUTHORITY[%d].NAME     = %s\n" % (i, `x[NAME]`))
        rv.append("AUTHORITY[%d].TYPE     = %d\n" % (i, x[TYPE]))
        rv.append("AUTHORITY[%d].CLASS    = %d\n" % (i, x[CLASS]))
        rv.append("AUTHORITY[%d].TTL      = %d\n" % (i, x[TTL]))
        rv.append("AUTHORITY[%d].RDLENGTH = %d\n" % (i, x[RDLENGTH]))
        rv.append("AUTHORITY[%d].RDATA    = %s\n" % (i, `x[RDATA]`))
    for i, x in enumerate(msg[ADDITIONAL]):
        rv.append("ADDITIONAL[%d].NAME     = %s\n" % (i, `x[NAME]`))
        rv.append("ADDITIONAL[%d].TYPE     = %d\n" % (i, x[TYPE]))
        rv.append("ADDITIONAL[%d].CLASS    = %d\n" % (i, x[CLASS]))
        rv.append("ADDITIONAL[%d].TTL      = %d\n" % (i, x[TTL]))
        rv.append("ADDITIONAL[%d].RDLENGTH = %d\n" % (i, x[RDLENGTH]))
        rv.append("ADDITIONAL[%d].RDATA    = %s\n" % (i, `x[RDATA]`))
    return "".join(rv)

def pretty_print(msg):
    print pretty_format(msg)

def build_name(labels):
    rv = []
    for lab in labels:
        ll = len(lab)
        if ll == 0:
            raise ValueError, "Zero-length label not allowed"
        elif ll > 63:
            raise ValueError, "Label too long (> 63 chars)"
        rv += [chr(ll), lab]
    rv.append(chr(0))
    return "".join(rv)

def build_int(n):
    return chr(n >> 8) + chr(n & 0xff)

def build_message(msg):
    # NB: bitfields can overflow if they exceed their limits
    if len(msg[QUESTION]) != msg[HEADER][QDCOUNT]:
        raise ValueError, "Incorrect QDCOUNT"
    if len(msg[ANSWER]) != msg[HEADER][ANCOUNT]:
        raise ValueError, "Incorrect ANCOUNT"
    if len(msg[AUTHORITY]) != msg[HEADER][NSCOUNT]:
        raise ValueError, "Incorrect NSCOUNT"
    if len(msg[ADDITIONAL]) != msg[HEADER][ARCOUNT]:
        raise ValueError, "Incorrect ARCOUNT"
    rv = []
    
    # Header
    rv.append(build_int(msg[HEADER][ID]))
    rv.append(chr(msg[HEADER][QR] << 7 | msg[HEADER][OPCODE] << 3 | msg[HEADER][AA] << 2 | msg[HEADER][TC] << 1 | msg[HEADER][RD]))
    rv.append(chr(msg[HEADER][Z] << 4 | msg[HEADER][RCODE]))
    rv.append(build_int(msg[HEADER][QDCOUNT]))
    rv.append(build_int(msg[HEADER][ANCOUNT]))
    rv.append(build_int(msg[HEADER][NSCOUNT]))
    rv.append(build_int(msg[HEADER][ARCOUNT]))
    
    # Question section
    for x in msg[QUESTION]:
        rv.append(build_name(x[QNAME]))
        rv.append(build_int(x[QTYPE]))
        rv.append(build_int(x[QCLASS]))
    
    # Answer, Authority and Additional sections
    for y in (msg[ANSWER], msg[AUTHORITY], msg[ADDITIONAL]):
        for x in y:
            rv.append(build_name(x[NAME]))
            rv.append(build_int(x[TYPE]))
            rv.append(build_int(x[CLASS]))
            rv.append(chr(x[TTL] >> 24) + chr(x[TTL] >> 16 & 0xff) + 
                      chr(x[TTL] >> 8 & 0xff) + chr(x[TTL] & 0xff))
            rv.append(build_int(x[RDLENGTH]))
            if len(x[RDATA]) != x[RDLENGTH]:
                raise ValueError, "Incorrect RDLENGTH"
            rv.append(x[RDATA])
    
    return "".join(rv)

def parse_name(s):
    f = StringIO(s)
    rv = []
    ll = ord(f.read(1))
    while ll != 0:
        lab = f.read(ll)
        if len(lab) != ll:
            raise ValueError, "name truncated"
        rv.append(lab)
        ll = ord(f.read(1))
    return tuple(rv)
